#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <math.h>
#include <vector>
#include "RtAudio.h"

using namespace std;

double PI = 3.141592653589793238462643383;
double g_time = 0;
unsigned int sampleRate = 44100;
bool input;

enum wave_type{SINE, SAW, PULSE, NOISE, IMPULSE, UNDEF};

class UnitGenerator;
typedef struct{
	int freq;
	float width;
	UnitGenerator *generator;
}generator_data;

/**
 * Only really used the classes/inheritance for practice-- not really helpful here.
 * Also, there are still some remnants from the first iteration, so it doesn't
 * fully take advantage of what it could do (like get rid of some globals)
 * but it still works, so . . . 
 */
class UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){}
};
class SineGenerator : public UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){
			for(int i = 0; i < nFrames; i++){
				buff[i] = sin(2*PI*data->freq*g_time/sampleRate);
				g_time += 1.0;
			}
		}
};
class SawtoothGenerator : public UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){
			int samps_per_period = (int)sampleRate/data->freq;
			for(int i = 0; i < nFrames; i++){
				int sampnum = (int)g_time % samps_per_period;
				double bound1 = ((data->width)*samps_per_period/2);
				double bound2 = bound1 + ((1-data->width) * samps_per_period);
				if(sampnum < bound1){
					buff[i] = (2*sampnum)/(data->width * samps_per_period);
				}else if(sampnum >= bound1 && sampnum < bound2){
					buff[i] = 1 - (2*sampnum)/((1 - data->width)*samps_per_period);
				}else if(sampnum >= bound2){
					sampnum -= (1-data->width) * samps_per_period + (((data->width) * samps_per_period)/2);
					buff[i] = ((2*sampnum)/(data->width * samps_per_period)) - 1.0;
				}
				g_time += 1.0;
			}
		}
};
class PulseGenerator : public UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){
			int samps_per_period = (int)sampleRate/data->freq;
			for(int i = 0; i < nFrames; i++){
				int sampnum = (int)g_time % samps_per_period;
				if(sampnum < (data->width * samps_per_period)){
					buff[i] = 1;
				}else{
					buff[i] = -1;
				}
				g_time += 1.0;
			}
		}
};
class NoiseGenerator : public UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){
			for(int i = 0; i < nFrames; i++){
				double randval = (((double)rand() / RAND_MAX) * 2) - 1.0;
				buff[i] = randval;
				g_time += 1.0;
			}
		}
};
class ImpulseGenerator : public UnitGenerator{
	public:
		virtual void genWave(double *buff, int nFrames, generator_data *data){
			int samps_per_period = (int)sampleRate/data->freq;
			for(int i = 0; i < nFrames; i++){
				int sampnum = (int)g_time % samps_per_period;
				if(sampnum == 0){
					buff[i] = 1;
				}else{
					buff[i] = 0;
				}
				g_time += 1.0;
			}
		}
};

int callback(void *outputBuffer, void *inputBuffer, unsigned int nFrames,
						 double streamTime, RtAudioStreamStatus status, void *userData){
	generator_data *data = (generator_data*)userData;
	double *buff = (double *)outputBuffer;
	data->generator->genWave(buff, nFrames, data);
	if(input){
		double *input = (double*)inputBuffer;
		for(int i = 0; i < nFrames; i++){
			buff[i] *= input[i];
		}
	}
	cout << "." << endl;
	return 0;
}

UnitGenerator *getGenerator(wave_type type){
	switch (type){
		case SINE:
			return new SineGenerator();
		case SAW:
			return new SawtoothGenerator();
		case PULSE:
			return new PulseGenerator();
		case NOISE:
			return new NoiseGenerator();
		case IMPULSE:
			return new ImpulseGenerator();
	}
	return NULL;
}

void printUsageInfoAndQuit(const char *errorMsg){
	cout << errorMsg << " Usage:" << endl;
	cout << "sig-gen [type] [frequency] [width] [flags]" << endl;
	cout << "[type]: --sine | --saw | --pulse | --noise | --impulse" << endl;
	cout << "[frequency]: (a number > 0, only applicable to some signal types)" << endl;
	cout << "[width]: pulse width (only applicable to some signal types)" << endl;
	cout << "[flags]: --input (microphone input)" << endl;
	exit(1);
}

bool isPossbleValue(char *typeArg, vector<char*> &vals){
	if(typeArg == NULL) return false;
	for(int i = 0; i < vals.size(); i++){
		if(strcmp(typeArg, vals[i]) == 0) return true;
	}
	return false;
}

wave_type getType(char *typeArg){
	if(typeArg == NULL) return UNDEF;
	if(strcmp(typeArg, "--sine") == 0) return SINE;
	if(strcmp(typeArg, "--saw") == 0) return SAW;
	if(strcmp(typeArg, "--pulse") == 0) return PULSE;
	if(strcmp(typeArg, "--noise") == 0) return NOISE;
	if(strcmp(typeArg, "--impulse") == 0) return IMPULSE;
	return UNDEF;
}

bool needsFreq(wave_type type){
	return type != NOISE;
}

bool needsWidth(wave_type type){
    switch(type){
        case SAW:
        case PULSE:
        case IMPULSE:
            return true;
        default:
            return false;
    }
}

int main (int argc, char * const argv[]) {
    RtAudio *audio = NULL;
	
	wave_type type;
	if((type = getType(argv[1])) == UNDEF){
		printUsageInfoAndQuit("Invalid Type Parameter.");
	}

	generator_data data2;
	data2.generator = getGenerator(type);

	if(needsFreq(type)){
        if(argc < 3) printUsageInfoAndQuit("Insufficient Number of Parameters.");
		int freq;
		sscanf(argv[2], "%d", &freq);
		data2.freq = freq;
		if(freq < 0) printUsageInfoAndQuit("Invalid Frequency (must be >= 0 and an integer).");
    }
    if(needsWidth(type)){
        if(argc < 4) printUsageInfoAndQuit("Insufficient Number of Parameters.");
		float width;
		sscanf(argv[3], "%f", &width);
		if(width < -1.0 || width > 1.0) printUsageInfoAndQuit("Invalid Width (must be between -1 and 1).");
		data2.width = width;
	}

	input = (strcmp(argv[argc - 1], "--input") == 0);

	unsigned int buffer_size = 512;
	
	try {
		audio = new RtAudio();
		
		RtAudio::StreamParameters *output_params = new RtAudio::StreamParameters();
		output_params->deviceId = audio->getDefaultOutputDevice();
		output_params->nChannels = 1;
		
		RtAudio::StreamParameters *input_params;
		if(input){
			input_params =  new RtAudio::StreamParameters();
			input_params->deviceId = audio->getDefaultInputDevice();
			input_params->nChannels = 1;
		}else{
			input_params = NULL;
		}
		
		RtAudio::StreamOptions *stream_options = new RtAudio::StreamOptions();
		stream_options->numberOfBuffers = 32;
		audio->openStream(output_params, input_params, RTAUDIO_FLOAT64, sampleRate, &buffer_size, &callback, &data2, stream_options);
		audio->startStream();
		while(true){
			usleep(10000);
		}
		audio->stopStream();
		audio->closeStream();
	}
	catch (RtError &err) {
		err.printMessage();
		exit(1);
	}
	
    return 0;
}
